"""
Fixtures for testing
"""

import pytest
import seamm_datastore

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


@pytest.fixture()
def session():
    some_engine = create_engine("sqlite:///:memory:")

    # create a configured "Session" class
    Session = sessionmaker(bind=some_engine)

    # create a Session
    sess = Session()
    return sess


@pytest.fixture(scope="function")
def connection():
    db = seamm_datastore.connect(initialize=True)
    users = db.get_users()
    for data in users:
        if data["username"] != "admin":
            user = data["username"]
            break

    db.login(username=user, password="default")
    return db


@pytest.fixture(scope="function")
def admin_connection():
    db = seamm_datastore.connect(initialize=True)
    db.login(username="admin", password="admin")
    return db


@pytest.fixture(scope="function")
def connection_nologin():
    db = seamm_datastore.connect(initialize=True)
    return db



"""
Marshmallow models for serialization and deserialization.

"""

import datetime

from marshmallow import fields
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from marshmallow_sqlalchemy.fields import Related, Nested

from seamm_datastore.database.models import Flowchart, Project, Job, User, Group, Role

#############################
#
# Custom Field
#
#############################


class LocalDateTime(fields.DateTime):
    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None
        data_format = self.format or self.DEFAULT_FORMAT
        value = value.replace(tzinfo=datetime.timezone.utc)
        value = value.astimezone()
        return value.strftime(data_format)


#############################
#
# Marshmallow Schema
#
#############################


class FlowchartSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Flowchart
        exclude = (
            "json",
            "owner_permissions",
            "group_permissions",
            "other_permissions",
            "sha256",
        )

    owner = Related("username")
    group = Related("name")


class ProjectSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Project
        exclude = ("owner_permissions", "group_permissions", "other_permissions")

    owner = Related("username")
    group = Related("name")


class JobSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Job
        exclude = (
            "flowchart",
            "owner_permissions",
            "group_permissions",
            "other_permissions",
        )

    owner = Related("username")
    group = Related("name")
    projects = Nested(
        ProjectSchema(
            only=(
                "name",
                "id",
            ),
            many=True,
        )
    )
    started = LocalDateTime(format="%Y-%m-%d %H:%M")
    finished = LocalDateTime(format="%Y-%m-%d %H:%M")
    submitted = LocalDateTime(format="%Y-%m-%d %H:%M")
    last_update = LocalDateTime(format="%Y-%m-%d %H:%M")


class UserSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = User
        exclude = ("password_hash",)


class GroupSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Group

    users = Nested(
        UserSchema(
            only=(
                "username",
                "id",
            ),
            many=True,
        )
    )


class RoleSchema(SQLAlchemyAutoSchema):
    class Meta:
        include_fk = True
        include_relationships = True
        model = Role



"""
Tests for build module
"""

import os

import seamm_datastore.util


def test_parse_flowchart_v2():
    this_file = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(this_file, "..", "data", "sample_flowchart_v2.flow")

    metadata, text = seamm_datastore.util.parse_flowchart(filepath)

    assert metadata["flowchart_version"] == 2.0
    assert (
        metadata["sha256_strict"]
        == "79d580b78559fe137872bcffe24aa7455e6c66fe260cf63e5edd3b3a1464e9c6"
    )

    assert text



"""
Util Functions and classes
"""

import json
import os
import re
from datetime import datetime, timezone

from dateutil import parser


def parse_flowchart(path):
    """
    Function for parsing information from flowchart

    Parameters
    ----------
    path: str
        The path to the flowchart.

    Returns
    -------
    metadata: dict
        A json containing flowchart information to be added to the database.
    """

    with open(path) as f:
        f.readline()
        version_info = " " + f.readline().split()[-1]
        text = f.read()

    if " 1." in version_info:
        metadata_pattern = None
        flowchart_pattern = re.compile("{.+}", re.DOTALL)
    elif " 2." in version_info:
        # Flowchart is version 2.
        metadata_pattern = re.compile("#metadata\n({.+?})\n#", re.DOTALL)
        flowchart_pattern = re.compile("#flowchart\n({.+})\n?#", re.DOTALL)
    else:
        # TODO Maybe raise custom exception. SEAMM Flowchart version error
        # Value Error for now
        raise ValueError

    # Handle the metadata
    if metadata_pattern:
        metadata = json.loads(metadata_pattern.findall(text)[0])
    else:
        metadata = {}

    # metadata as a json, flowchart (json) as text.
    metadata["flowchart_version"] = float(version_info)
    flowchart = flowchart_pattern.findall(text)[0]

    return metadata, flowchart


def parse_job_data(job_data_json):
    """Parse job_data.json at path"""
    directory = job_data_json["working directory"]
    job_data = {
        "path": directory,
        "title": str(
            job_data_json["title"]
            if job_data_json["title"]
            else os.path.basename(directory)
        ),
        "project_names": job_data_json["projects"],
        "status": job_data_json["state"],
        "id": job_data_json["job id"],
    }

    if "end time" in job_data_json:
        try:
            job_data["finished"] = datetime.fromisoformat(job_data_json["end time"])
        except Exception:
            job_data["finished"] = parser.parse(job_data_json["end time"]).astimezone(
                timezone.utc
            )

    if "start time" in job_data_json:
        try:
            job_data["started"] = datetime.fromisoformat(job_data_json["start time"])
        except Exception:
            job_data["started"] = parser.parse(job_data_json["start time"]).astimezone(
                timezone.utc
            )

    if "submitted time" in job_data_json:
        job_data["submitted"] = datetime.fromisoformat(job_data_json["submitted time"])
    elif "started" in job_data:
        job_data["submitted"] = job_data["started"]

    return job_data


class LoginRequiredError(Exception):
    pass


class NotAuthorizedError(Exception):
    pass



"""
Tests for building/importing datastore
"""

import os


def test_build(connection):

    loc = os.path.abspath(os.path.dirname(__file__))
    added_jobs, added_projects = connection.import_datastore(
        os.path.join(loc, "..", "data", "Projects")
    )

    assert len(added_jobs) == 2
    assert len(added_projects) == 1, added_projects



"""
seamm_datastore
The database models for the seamm datastore
"""

# Imports - alias class to "connect".
import seamm_datastore.api  # noqa: F401
from .connect import SEAMMDatastore as connect

__all__ = ["connect"]

# Handle versioneer
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



"""
Table models for SEAMM datastore SQLAlchemy database.
"""

from datetime import datetime
from werkzeug.security import generate_password_hash, check_password_hash

from sqlalchemy import (
    Column,
    Integer,
    String,
    DateTime,
    ForeignKey,
    Table,
    Text,
    JSON,
    Float,
)
from sqlalchemy.orm import relationship

from sqlalchemy.ext.declarative import declarative_base

# Patched flask authorize
from seamm_datastore.flask_authorize_patch import (
    AccessControlPermissionsMixin,
    generate_association_table,
)

# The default is sqlalchemy unless we have
# the dashboard installed and a db with
# a bound engine.
try:
    import sys

    assert "seamm_dashboard" in sys.modules
    from seamm_dashboard import db

    Base = db.Model
except AssertionError:
    Base = declarative_base()

#############################
#
# SQLAlchemy Models
#
#############################

# Authentication Mapping Tables for Access Control
UserProjectMixin = generate_association_table("User", "Project")
GroupProjectMixin = generate_association_table("Group", "Project")


class UserProjectAssociation(Base, UserProjectMixin):
    pass


class GroupProjectAssociation(Base, GroupProjectMixin):
    pass


user_group = Table(
    "user_group",
    Base.metadata,
    Column("user", Integer, ForeignKey("users.id"), primary_key=True),
    Column("group", Integer, ForeignKey("groups.id"), primary_key=True),
)

user_role = Table(
    "user_role",
    Base.metadata,
    Column("user", Integer, ForeignKey("users.id"), primary_key=True),
    Column("role", Integer, ForeignKey("roles.id"), primary_key=True),
)


flowchart_project = Table(
    "flowchart_project",
    Base.metadata,
    Column("flowchart", String(32), ForeignKey("flowcharts.id"), primary_key=True),
    Column("project", Integer, ForeignKey("projects.id"), primary_key=True),
)

job_project = Table(
    "job_project",
    Base.metadata,
    Column("job", Integer, ForeignKey("jobs.id"), primary_key=True),
    Column("project", Integer, ForeignKey("projects.id"), primary_key=True),
)


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    username = Column(String, unique=True, nullable=False)
    first_name = Column(String)
    last_name = Column(String)
    email = Column(String, unique=True)
    password_hash = Column(String)
    added = Column(DateTime, nullable=False, default=datetime.utcnow)
    status = Column(String, default="active")

    roles = relationship("Role", secondary=user_role, back_populates="users")
    groups = relationship("Group", secondary=user_group, back_populates="users")

    @property
    def password(self):
        raise AttributeError("password is not a readable attribute")

    @password.setter
    def password(self, password):
        self.password_hash = generate_password_hash(password)

    def verify_password(self, password):
        return check_password_hash(self.password_hash, password)


class Group(Base):
    __tablename__ = "groups"

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, nullable=False)

    users = relationship("User", secondary=user_group, back_populates="groups")


class Role(Base):
    __tablename__ = "roles"

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, nullable=False)

    users = relationship("User", secondary=user_role, back_populates="roles")


class Flowchart(Base, AccessControlPermissionsMixin):
    __tablename__ = "flowcharts"

    id = Column(Integer, nullable=False, primary_key=True)
    sha256 = Column(String(75), nullable=True)
    sha256_strict = Column(String(75), nullable=True)
    path = Column(String, nullable=True)
    flowchart_version = Column(Float, nullable=True, unique=False)
    doi = Column(Text, nullable=True)
    title = Column(String(100), nullable=True)
    description = Column(Text, nullable=True)
    creators = Column(JSON, nullable=True)
    keywords = Column(JSON, nullable=True)
    json = Column(JSON, nullable=False)

    jobs = relationship("Job", back_populates="flowchart", lazy=True)
    projects = relationship(
        "Project", secondary=flowchart_project, back_populates="flowcharts"
    )

    def __repr__(self):
        return f"Flowchart(id={self.id}, description={self.description}, path={self.path})"  # noqa: E501


class Job(Base, AccessControlPermissionsMixin):
    __tablename__ = "jobs"

    id = Column(Integer, primary_key=True)
    flowchart_id = Column(String(32), ForeignKey("flowcharts.id"))
    title = Column(String, nullable=True)
    description = Column(Text, nullable=True)
    path = Column(String, unique=True)
    submitted = Column(DateTime, nullable=False, default=datetime.utcnow)
    started = Column(DateTime, nullable=True)
    finished = Column(DateTime, nullable=True)
    status = Column(String, nullable=False, default="imported")
    last_update = Column(
        DateTime, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow
    )
    parameters = Column(JSON, nullable=True)

    flowchart = relationship("Flowchart", back_populates="jobs")
    projects = relationship("Project", secondary=job_project, back_populates="jobs")

    def __repr__(self):
        return f"Job(path={self.path}, flowchart_id={self.flowchart}, submitted={self.submitted})"  # noqa: E501


class Project(Base, AccessControlPermissionsMixin):
    __tablename__ = "projects"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    description = Column(String(1000), nullable=True)
    path = Column(String, unique=True)

    flowcharts = relationship(
        "Flowchart", secondary=flowchart_project, back_populates="projects"
    )
    jobs = relationship("Job", secondary=job_project, back_populates="projects")

    def __repr__(self):
        return f"Project(name={self.name}, path={self.path}, description={self.description})"  # noqa: E501



"""
Unit and regression test for the seamm_datastore package.
"""

# Import package, test suite, and other packages as needed
from pathlib import Path
import pprint
import pytest

from dateutil import parser


def test_connected(admin_connection):
    assert admin_connection.current_user().username == "admin"
    assert admin_connection.default_project == "default"


def test_connection_logout(connection):
    connection.logout()
    assert connection.current_user() is None


def test_add_job_error(connection):
    from seamm_datastore.util import LoginRequiredError

    connection.logout()
    with pytest.raises(LoginRequiredError):
        connection.add_job({"title": "fake job"})


def test_get_users(connection):
    users = connection.get_users()
    if len(users) != 2:
        pprint.pprint(users)
    assert len(users) == 2


def test_add_job(connection):
    path = (
        Path(__file__)
        / ".."
        / ".."
        / "data"
        / "Projects"
        / "sample_project1"
        / "Job_000093"
    )
    path = path.expanduser().resolve()

    # connection.add_job(job1_data)
    connection.add_job(
        1,
        str(path / "flowchart.flow"),
        project_names=["default"],
        path=str(path),
        title="test job",
        description="description of the job",
        submitted=parser.parse("2016-08-29T09:12:33.000000+00:00"),
        started=parser.parse("2016-08-29T09:12:34.000000+00:00"),
        finished=parser.parse("2016-08-29T09:13:34.000000+00:00"),
        status="finished",
    )

    # Retrieve job
    assert len(connection.get_jobs()) == 1



"""
seamm_datastore
The database models for the seamm datastore
"""
import sys
from setuptools import setup, find_packages
import versioneer

short_description = __doc__.splitlines()[1]

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except Exception:
    long_description = "\n".join(short_description[2:])

with open('requirements_install.txt') as fd:
    requirements = fd.read()


setup(
    # Self-descriptive entries which should always be present
    name='seamm_datastore',
    author='Jessica A. Nash (The Molecular Sciences Software Institute)',
    author_email='janash@vt.edu',
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license='BSD-3-Clause',
    url='https://github.com/molssi-seamm/seamm_datastore',

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,

    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,

    # Required packages, pulls from pip if needed; do not use for Conda
    # deployment
    install_requires=requirements,

    test_suite='tests',

    # Valid platforms your code works on, adjust to your flavor
    platforms=['Linux',
               'Mac OS-X',
               'Unix'],

    # Manual control if final package is compressible or not, set False to
    # prevent the .egg from being made
    # zip_safe=False,

    keywords=['SEAMM', 'datastore', 'dashboard'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



"""
Functions which take a session and add or retrieve data to the db
"""
import datetime
import pprint  # noqa: F401

from seamm_datastore.util import parse_flowchart


def _build_query(filter, obj):
    q = []
    for key, value in filter.items():
        q.append(getattr(obj, key).in_(value))
    return q


def add_flowchart(session, flowchart_info):
    from seamm_datastore.database.models import Flowchart

    try:
        flowchart = Flowchart.query.filter_by(
            sha256_strict=flowchart_info["sha256_strict"]
        ).one_or_none()
    except KeyError:
        try:
            flowchart = Flowchart.query.filter_by(id=flowchart_info["id"]).one_or_none()
        except KeyError:
            flowchart = None

    if flowchart:
        raise ValueError(f"Flowchart already in datastore. ID: {flowchart.id}")

    new_flowchart = Flowchart(**flowchart_info)

    session.add(new_flowchart)
    session.commit()

    return new_flowchart


def add_group(
    session,
    name,
    as_json=False,
    current_user=None,
):
    """
    Add a project to the database.

    Parameters
    ----------
    session : sqlalchemy.Session
        The session used to access the database.
    name : str
        The name of the project, used for display and directory name.
    as_json : bool = False
        If True, return the json description of the project; otherwise, the project id.
    current_user : str or User = None
        The user currently logged in. Not used.

    Returns
    -------
    json or Group
        The json describing the new project, or the Group object, deoending on
        "as_json".
    """
    from seamm_datastore.database.models import Group
    from seamm_datastore.database.schema import GroupSchema

    # Check that the group doesn't already exist.
    group = Group.query.filter_by(name=name).one_or_none()
    if group is not None:
        raise ValueError(f"Group '{group}' already exists.")

    group = Group(name=name)
    session.add(group)
    session.commit()

    # Return the json or id as requested.
    if as_json:
        group_schema = GroupSchema()
        return group_schema.dump(group)
    else:
        return group


def add_job(
    session,
    job_id,
    flowchart_filename,
    project_names=["default"],
    path=None,
    title="",
    description="",
    parameters={},
    submitted=datetime.datetime.now(datetime.timezone.utc),
    started=None,
    finished=None,
    status="submitted",
    as_json=False,
    current_user=None,
):
    """Submit a job to the datastore.

    This method requires a user to be logged in and to have appropriate permissions
    for the project.

    Parameters
    ----------
    session : sqlalchemy.Session
        The session used to access the database.
    job_id : int
        The id of the job, an integer > 0.
    flowchart_filename : str or pathlib.Path
        The path to the file containing the flowchart.
    project_names : [str]
        A list of projects that the flowchart belongs to.
    path : str
        The directory path for the job.
    title : str = ""
        The title of the job, used for display.
    description : str = ""
        A longer, textual description of the job.
    parameters : dict = {}
        The command-line parameters for the job.
    submitted : datetime.datetime = now()
        When the job was submitted as a datetime object. Defaults to now in UTC.
    started : datetime.datetime = None
        When the job was started, if it was. Preferably in UTC.
    finished : datetime.datetime = None
        When the job finished, if it has. Preferably in UTC.
    status : str = "submitted"
        The status of the job: "submitted", "running", "finished", "error", etc.
    as_json : bool = False
        If true return the job data asjson, otehrwise return the Job object.
    current_user : str or User = None
        The user currently logged in. Not used.

    Returns
    -------
    json or Job
        The json of the job data, or the Job object, depending on "as_json".
    """
    from seamm_datastore.database.models import Job, Project
    from seamm_datastore.database.schema import JobSchema

    # Check if this job already exists
    try:
        job = Job.query.filter_by(id=job_id).one_or_none()
    except KeyError:
        job = None
    if job:
        raise ValueError(f"Job with ID {job_id} already found in the database")

    # Get the ids for the projects
    projects = [Project.query.filter_by(name=x).one_or_none() for x in project_names]
    projects = [project for project in projects if project]

    if not projects:
        tmp = "', '".join(project_names)
        raise NameError(
            "Projects listed for this job not found in database."
            f"\nPlease check your project names: '{tmp}'"
        )

    # Check the permissions of the projects to see if we can add a job to them
    allowed_projects = Project.query.filter(Project.authorized("update")).all()
    for project in projects:
        # if project not in Project.query.filter(Project.authorized("update")).all():
        if project not in allowed_projects:
            raise RuntimeError(
                f"You are not authorized to add jobs to {project} project."
            )

    # Handle the flowchart - we'll only want to add it if we're adding the job.
    fl_data, fl = parse_flowchart(flowchart_filename)
    fl_data["json"] = fl

    try:
        flowchart = add_flowchart(session, fl_data)
    except ValueError as e:
        from seamm_datastore.database.models import Flowchart

        id = int(e.args[0].split(":")[1])
        flowchart = Flowchart.query.filter_by(id=id).one()

    # Now add the job and update the database
    new_job = Job(
        id=job_id,
        title=title,
        description=description,
        path=path,
        flowchart=flowchart,
        projects=projects,
        parameters=parameters,
        status=status,
        submitted=submitted,
        started=started,
        finished=finished,
    )
    session.add(new_job)
    session.commit()

    # Return the job as json or an object.
    if as_json:
        return JobSchema().dump(new_job)
    else:
        return new_job


def add_project(
    session,
    name,
    description="",
    path=None,
    owner=None,
    group=None,
    as_json=False,
    current_user=None,
):
    """
    Add a project to the database.

    Parameters
    ----------
    session : sqlalchemy.Session
        The session used to access the database.
    name : str
        The name of the project, used for display and directory name.
    description : str
        A textual description of the project.
    path : str = None
        The path on disk to the project files.
    owner : str or User = None
        An optional user to own the project. Defaults to the currently logged in user.
    group : str or Group = None
        The group for the project. Defaults to user's primary group.
    as_json : bool = False
        If True, return the json description of the project; otherwise, the project id.
    current_user : str or User = None
        The user currently logged in.

    Returns
    -------
    json or Project
        The json describing the new project, or the Project object, deoending on
        "as_json".
    """
    from seamm_datastore.database.models import Project, User, Group
    from seamm_datastore.database.schema import ProjectSchema, UserSchema

    # Check that the project doesn't already exist.
    project = Project.query.filter_by(name=name).one_or_none()
    if project is not None:
        raise ValueError(f"Project {project} already found in the database")

    # Sort out the user and get as a User object.
    if owner is None:
        if current_user is None:
            raise ValueError("The owner is required for adding a project.")
        else:
            owner = current_user
    if isinstance(owner, str):
        owner_id = User.query.filter_by(username=owner).one()
        owner = owner_id

    # Get the group as a Group object. The default is the user's first group.
    if group is None:
        owner_info = UserSchema().dump(owner)
        group = Group.query.filter_by(id=owner_info["groups"][0]).one()
    elif isinstance(group, str):
        group_id = Group.query.filter_by(name=group).one()
        group = group_id

    # Create the project and sotr in the database
    project = Project(
        name=name, description=description, path=path, owner=owner, group=group
    )
    session.add(project)
    session.commit()

    # Return the json or id as requested.
    if as_json:
        project_schema = ProjectSchema()
        return project_schema.dump(project)
    else:
        return project


def add_user(
    session,
    username,
    password,
    first_name=None,
    last_name=None,
    email=None,
    roles=["user"],
    groups=["staff"],
    as_json=False,
    current_user=None,
):
    """
    Add a new user to the database.

    Parameters
    ----------
    session : sqlalchemy.Session
        The session used to access the database.
    username : str
        The username that identifies the user.
    password : str
        The secret password for the user.
    first_name : str = None
        The user's first (given) name.
    last_name : str = None
        The user's last (family) name.
    email : str = None
        The user's principal email address.
    roles : [str] = ["user"]
        A list of roles for the user. Defaults to just "user".
    groups : [str] = ["staff"]
        A list of groups that the user belongs to. Defaults to "staff".
    as_json : bool = False
        If True, return the json description of the user; otherwise, the user id.
    current_user : str or User = None
        The user currently logged in. Not used.
    """
    from seamm_datastore.database.models import Group, Role, User
    from seamm_datastore.database.schema import UserSchema

    # Check if the user already exists.
    user = User.query.filter_by(username=username).one_or_none()
    if user:
        raise ValueError(f"User {user} already found in the database")

    # Create the new user and store in the database
    new_user = User(
        username=username,
        password=password,
        first_name=first_name,
        last_name=last_name,
        email=email,
    )
    for role_name in roles:
        role = Role.query.filter_by(name=role_name).one()
        new_user.roles.append(role)
    for group_name in groups:
        group = Group.query.filter_by(name=group_name).one()
        new_user.groups.append(group)
    session.add(new_user)
    session.commit()

    # Return the json or user, as requested.
    if as_json:
        return UserSchema().dump(new_user)
    else:
        return new_user


def finish_job(
    session,
    job_id,
    finish_time,
    status="finished",
    as_json=False,
    current_user=None,
):
    """Set the status and time that the job finished.

    Parameters
    ----------
    session : session object
        The SQLAlchemy session
    job_id : int
        The ID of the job, eg. 209
    finish_time : datetime.datetime
        The UTC time when the job finished.
    status : str
        The status, such as "error" or the default, "finished"
    as_json : bool = False
        Ignored
    current_user : str or User = None
        Ignored

    Returns
    -------
    bool
        True if the finish time was successfully set, False otherwise.
    """
    from seamm_datastore.database.models import Job

    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        return False
    else:
        job.finished = finish_time
        job.status = status
        session.commit()
        return True


def get_flowcharts(_=None, as_json=False, current_user=None):
    from seamm_datastore.database.models import Flowchart

    flowcharts = Flowchart.query.filter(Flowchart.authorized("read")).all()

    if as_json:
        from seamm_datastore.database.schema import FlowchartSchema

        flowcharts = FlowchartSchema(many=True).dump(flowcharts)

    return flowcharts


def get_groups(_=None, as_json=False, current_user=None):
    from seamm_datastore.database.models import Group

    groups = Group.query.all()

    if as_json:
        from seamm_datastore.database.schema import GroupSchema

        groups = GroupSchema(many=True).dump(groups)

    return groups


def get_job(session, id, as_json=False):

    from seamm_datastore.database.models import Job, Project
    from .util import NotAuthorizedError

    authorized_job = Job.query.filter(Job.authorized("read"), Job.id == id)
    project_job = Project.query.with_entities(Job).filter(
        Project.authorized("read"), Job.id == id
    )

    project_job = Job.query.filter(
        Job.projects.any(Project.authorized("read")), Job.id == id
    )

    # Do we get the job?
    job = authorized_job.union(project_job).one_or_none()

    if job is None and Job.query.get(id):
        raise NotAuthorizedError

    if as_json:
        from seamm_datastore.database.schema import JobSchema

        job = JobSchema().dump(job)

    return job


def get_jobs(_=None, as_json=False, limit=None, offset=None, count=False, status=None):

    from seamm_datastore.database.models import Job, Project

    # Get the jobs where the user has read permission
    if status is None:
        authorized_jobs = Job.query.filter(Job.authorized("read"))
    else:
        authorized_jobs = Job.query.filter(Job.authorized("read"), Job.status == status)

    # Get jobs from projects where user has read permission
    if status is None:
        project_jobs = Job.query.filter(Job.projects.any(Project.authorized("read")))
    else:
        project_jobs = Job.query.filter(
            Job.projects.any(Project.authorized("read")), Job.status == status
        )

    # Find union of these queries
    jobs = authorized_jobs.union(project_jobs)

    # Continue building
    if limit is not None:
        jobs = jobs.limit(limit)
    if offset is not None:
        jobs = jobs.offset(offset)

    if count is True:
        jobs = jobs.count()
    else:
        jobs = jobs.all()

        if as_json:
            from seamm_datastore.database.schema import JobSchema

            jobs = JobSchema(many=True).dump(jobs)

    return jobs


def get_projects(
    session=None,
    action="read",
    filter=None,
    limit=None,
    offset=None,
    count=False,
    as_json=False,
    current_user=None,
):
    """Get the projects in the datastore.

    Parameters
    ----------
    session : Session object = None
        The SQLAlchemy session or equivalent. Not used.
    action : str = "read"
        Whether to get readable ("read") of writeable ("update") projects.
    filter : dict = None
        A dictionary of query conditions for filtering the results.
    limit : int = None
        The maximum number of records to return.
    offset : int = None
        Return the records starting at offset (0-based). Used with limit to page through
        records.
    count : bool = False
        If true, return the number of records returned by the query.
    as_json : bool = False
        If true return json for an array of dictionaries describing the projects.
        If false, return just the list of project names.
    current_user : str or User = None
        The user currently logged in. Not used.

    Returns
    -------
    int, [str], or json
        The integer count of records if "count" is True.
        A list of project names if "as_json" is False.
        And array of dictionaries as json if "as_json" is True.
    """
    from seamm_datastore.database.models import Project

    if filter is None:
        query = Project.query.filter(Project.authorized(action))
    else:
        # Build a query
        b = _build_query(filter, Project)
        # Pass the query. Use unpacking operator to unpack generator.
        projects = Project.query.filter(
            Project.authorized(action), *(a for a in b)
        ).all()

    if limit is not None:
        query = query.limit(limit)
    if offset is not None:
        query = query.offset(offset)

    if count:
        result = query.count()
    else:
        projects = query.all()
        if as_json:
            from seamm_datastore.database.schema import ProjectSchema

            result = ProjectSchema(many=True).dump(projects)
        else:
            result = [p.name for p in projects]

    return result


def get_users(_=None, as_json=False, current_user=None):
    from seamm_datastore.database.models import User

    users = User.query.all()

    if as_json:
        from seamm_datastore.database.schema import UserSchema

        users = UserSchema(many=True).dump(users)

    return users


def list_projects(
    _=None,
    action="read",
    filter=None,
    limit=None,
    offset=None,
    count=False,
    as_json=False,
    current_user=None,
):
    """Get a list of projects in the datastore.

    Parameters
    ----------
    session : Session object = None
        The SQLAlchemy session or equivalent. Not used.
    action : str = "read"
        Whether to get readable ("read") of writeable ("update") projects.
    filter : dict = None
        A dictionary of query conditions for filtering the results.
    limit : int = None
        The maximum number of records to return.
    offset : int = None
        Return the records starting at offset (0-based). Used with limit to page through
        records.
    count : bool = False
        If true, return the number of records returned by the query.
    as_json : bool = False
        Ignored
    current_user : str or User = None
        Ignored

    Returns
    -------
    int, [str]
        The integer count of records if "count" is True, otherwise a list of project
        names.
    """
    return get_projects(
        None,
        action=action,
        filter=filter,
        limit=limit,
        offset=offset,
        count=count,
        as_json=None,
    )



"""
Automatic import of projects and jobs from directories.
"""

import json
import os
from pathlib import Path

from seamm_datastore import api

from seamm_datastore.util import parse_job_data


def _build_initial(session, default_project):
    """Build the initial database"""

    from seamm_datastore.database.models import Role, Group, User, Project

    # Create roles
    role_names = ["user", "group manager", "admin"]
    for role_name in role_names:
        role = Role(name=role_name)
        session.add(role)
        session.commit()

    # Create default admin group
    admin_group = Group(name="admin")
    session.add(admin_group)
    session.commit()

    # Create default admin user.s
    admin_role = session.query(Role).filter_by(name="admin").one()
    admin_user = User(username="admin", password="admin", roles=[admin_role])
    admin_user.groups.append(admin_group)

    # Create a user and group with the same information as user running
    try:
        item = Path.home()
        username = item.owner()
        group_name = item.group()
    except NotImplementedError:
        # This will occur on Windows
        import os

        username = os.environ["USERNAME"]
        # Just a default group name.
        group_name = "staff"

    group = Group(name=group_name)

    password = "default"
    user = User(username=username, password=password, roles=[admin_role])
    user.groups.append(group)

    # Admin user needs to be part of all groups.
    admin_user.groups.append(group)

    session.add(admin_user)
    session.add(admin_role)
    session.add(admin_group)

    session.add(user)

    # Create a default project
    project = Project(name=default_project, owner=user, group=group)
    session.add(project)
    session.commit()


def import_datastore(session, location, as_json=True):
    """Import all the projects and jobs at <location>.

    Parameters
    ----------
    session : SQLAlchemy or flask session
    location : str or path
        The location to check for jobs or projects. Usually the projects directory in a
        datastore.

    Returns
    -------
    (n_projects, n_jobs) : int, integer
        The number of projects and jobs added to the database.
    """

    from seamm_datastore.database.models import Project

    jobs = []
    project_names = []

    # Get directory contents of file path
    for folder in os.listdir(location):
        potential_project = os.path.join(location, folder)

        # If item is a directory, it may contain jobs.
        # We are going to be taking the project names
        # from the job_data.json
        if os.path.isdir(potential_project):
            project_name = os.path.basename(potential_project)
            item = Path(potential_project)
            try:
                group = item.group()
                username = item.owner()
            except NotImplementedError:
                username = os.environ["USERNAME"]
                # Just a default group name.
                group = "staff"

            project_data = {
                "owner": username,
                "group": group,
                "name": project_name,
                "path": potential_project,
            }

            try:
                api.add_project(
                    session,
                    name=project_name,
                    path=potential_project,
                    owner=username,
                    group=group,
                    as_json=as_json,
                )
                project_names.append(project_data["name"])
            except Exception:
                # Project exists, we don't need to add it.
                # Pass here because we should still try importing jobs.
                session.rollback()

            for potential_job in os.listdir(potential_project):
                potential_job = os.path.join(potential_project, potential_job)

                if os.path.isdir(potential_job):

                    # Check for job_data.json - has to have this to be job
                    check_path = os.path.join(potential_job, "job_data.json")

                    if os.path.exists(check_path):
                        with open(check_path, "r") as fd:
                            lines = fd.read().splitlines()
                        # Old files may not have a header line
                        if lines[0][0] == "{":
                            text = "\n".join(lines)
                        else:
                            text = "\n".join(lines[1:])
                        job_data_json = json.loads(text)
                        job_data = parse_job_data(job_data_json)

                        try:
                            job = api.add_job(
                                session,
                                job_data["id"],
                                potential_job + "/flowchart.flow",
                                project_names=job_data["project_names"],
                                path=potential_job,
                                title=job_data["title"],
                                description=job_data.get("description", ""),
                                submitted=job_data.get("submitted", None),
                                started=job_data.get("started", None),
                                finished=job_data.get("finished", None),
                                status=job_data["status"],
                                as_json=as_json,
                            )
                            jobs.append(job)
                        except Exception as e:
                            print(f"Exception --> {str(e)}")
                            # Job has already been added.
                            session.rollback()

    # retrieve projects now that all the jobs have been added.
    projects = Project.query.filter(Project.name.in_(project_names)).all()

    return jobs, projects



"""
Unit tests for projects in the seamm_datastore package.
"""
import pprint

import pytest


@pytest.fixture(scope="function")
def tester(connection_nologin):
    """Create a connection with user "tester"."""
    conn = connection_nologin
    conn.login("admin", "admin")
    conn.add_group("test")
    conn.add_user("tester", "default", groups=["test"])
    conn.logout()
    conn.login("tester", "default")

    return conn


@pytest.fixture(scope="function")
def many(tester):
    """Create a connection with many projects."""
    conn = tester
    for i in range(1, 101):
        name = f"project {i}"
        path = f"pdir_{i}"
        conn.add_project(name, owner="tester", group="test", path=path)

    return conn


def test_get_(connection):
    projects = connection.get_projects()
    if len(projects) != 1:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 1


def test_get_nologin(connection_nologin):
    projects = connection_nologin.get_projects()
    if len(projects) != 0:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 0


def test_add_many(tester):
    correct = []
    for i in range(1, 11):
        name = f"project {i}"
        path = f"pdir_{i}"
        tester.add_project(name, owner="tester", group="test", path=path)
        correct.append(name)
    projects = tester.list_projects()
    if len(projects) != 10 or projects != correct:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 10
    assert projects == correct


def test_many(many):
    projects = many.list_projects()
    if len(projects) != 100:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 100


def test_limit(many):
    projects = many.list_projects(limit=10)
    if len(projects) != 10:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 10


def test_offset(many):
    correct = ["project 2", "project 3", "project 4"]
    projects = many.list_projects(limit=3, offset=1)
    if len(projects) != 3 or projects != correct:
        pprint.pprint(f"{projects=}")
    assert len(projects) == 3
    assert projects == correct


def test_count(many):
    count = many.list_projects(count=True)
    assert count == 100



"""
Class and functions for connection to database.
"""

import os

from functools import wraps
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from flask_authorize import Authorize

import seamm_datastore.database.build
from seamm_datastore import api
from .util import LoginRequiredError


def manage_session(function):
    """Decorator for closing sqlalchemy sessions when attached to SEAMMDatastore
    class.
    """

    def _wrap_method(method):
        def _manage_session(self, *args, **kwargs):
            with session_scope(self.Session) as session:
                ret = function(session, as_json=True, *args, **kwargs)
            return ret

        return _manage_session

    return _wrap_method


@contextmanager
def session_scope(session):
    """Provide a transactional scope around a series of operations."""
    session = session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def login_required(method):
    """Decorator for actions requiring current user"""

    @wraps(method)
    def _check_user(self, *args, **kwargs):
        if not self.current_user():
            raise LoginRequiredError

        else:
            ret = method(self, *args, **kwargs, current_user=self.current_user())
        return ret

    return _check_user


class current_app:
    """Fake current_app"""

    def __init__(self, config):
        self.config = config
        self.extensions = ["sqlalchemy"]


class SEAMMDatastore:
    def __init__(
        self,
        database_uri: str = "sqlite:///:memory:",
        initialize: bool = False,
        permissions: dict = None,
        username: str = None,
        password: str = None,
        datastore_location: str = None,
        default_project: str = "default",
    ):

        # Default permissions
        if permissions is None:
            permissions = {
                "owner": ["read", "update", "delete"],
                "group": ["read", "update"],
                "world": [],
            }

        global fake_app
        fake_app = current_app(
            config={
                "AUTHORIZE_DEFAULT_PERMISSIONS": permissions,
                "AUTHORIZE_MODEL_PARSER": "table",
                "AUTHORIZE_IGNORE_PROPERTY": "__check_access__",
            }
        )

        # Set datastore location
        if datastore_location is None:
            self.datastore_location = os.getcwd()
        else:
            self.datastore_location = datastore_location

        # Create engine and session
        self.engine = create_engine(database_uri)

        self.Session = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )

        from seamm_datastore.database.models import Base, Project

        if initialize:
            Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)
        Base.query = self.Session.query_property()

        # Set up the database.
        if initialize:
            seamm_datastore.database.build._build_initial(
                self.Session(), default_project
            )

            from seamm_datastore.database.models import Group, Role

            group = Group.query.filter_by(name="admin").one()
            admin_role = Role.query.filter_by(name="admin").one()

            # Log in the admin user and create specified user
            if username:
                self.login("admin", "admin")
                self.add_user(username, password, roles=[admin_role], groups=[group])
                self.logout()

        if username:
            self.login(username, password)
        else:
            self._user = None

        # Current user has to be bound before we can add anything else to be database.
        self.authorize = Authorize(current_user=self.current_user)

        # Now handle the project
        project = Project.query.filter_by(name=default_project).one_or_none()
        if not project:
            raise ValueError("Invalid project name given for default.")

        self.default_project = default_project

    def login(self, username, password):
        from seamm_datastore.database.models import User

        user = User.query.filter_by(username=username).one()

        if not user.verify_password(password):
            raise ValueError("Login unsuccessful. Check username and password.")

        self._user = username

    def logout(self):
        self._user = None

    # Seems like a good place for @property, but can't use because flask authorize
    # requires this to be callable.
    def current_user(self):
        from seamm_datastore.database.models import User

        if self._user:
            user = User.query.filter_by(username=self._user).one_or_none()
        else:
            user = None
        return user

    @manage_session(api.add_flowchart)
    def add_flowchart(self, *args, **kwargs):
        pass

    @manage_session(api.add_group)
    def add_group(self, *args, **kwargs):
        pass

    @login_required
    @manage_session(api.add_job)
    def add_job(self, *args, **kwargs):
        pass

    @login_required
    @manage_session(api.add_project)
    def add_project(self, *args, **kwargs):
        pass

    @login_required
    @manage_session(api.add_user)
    def add_user(self, *args, **kwargs):
        pass

    @login_required
    @manage_session(api.finish_job)
    def finish_job(self, *args, **kwargs):
        pass

    @manage_session(api.get_flowcharts)
    def get_flowcharts(self, *args, **kwargs):
        pass

    @manage_session(api.get_groups)
    def get_groups(self, *args, **kwargs):
        pass

    @manage_session(api.get_jobs)
    def get_jobs(self, *args, **kwargs):
        pass

    @manage_session(api.get_projects)
    def get_projects(self, *args, **kwargs):
        pass

    @manage_session(api.get_users)
    def get_users(self, *args, **kwargs):
        pass

    @manage_session(api.list_projects)
    def list_projects(self, *args, **kwargs):
        pass

    @manage_session(seamm_datastore.database.build.import_datastore)
    def import_datastore(self, *args, **kwargs):
        pass
